/*
 * tabmix.js
 *
 * code by Hemiola SUN
 */

const __TM_Branch = "extensions.tabmix.";
const __TM_Prefs  = Services.prefs.getBranch(__TM_Branch);

/*
 * Function to catch changes of preferences and update existing tabs
 */
var TMPrefObserver =
{
  register: function () {
    Services.prefs.addObserver(__TM_Branch, this, false);
    Services.prefs.addObserver("browser.tabs.", this, false);
  },

  unregister: function () {
    Services.prefs.removeObserver(__TM_Branch, this, false);
    Services.prefs.removeObserver("browser.tabs.", this, false);
  },

  observe: function (aSubject, aTopic, aData) {
    if (aTopic == "nsPref:changed") 
      TabMix.onUpdate();
  }
}

/*
 *========
 * TabMix
 *========
 */
var TabMix = 
{
  _Ss : Cc["@mozilla.org/browser/sessionstore;1"].getService(Ci.nsISessionStore),
      
  _ClickTab_Null         : 0,
  _ClickTab_OpenNewTab   : 1,
  _ClickTab_RemoveTab    : 2,
  _ClickTab_DuplicateTab : 3,
  _ClickTab_ReloadTab    : 4,
  _ClickTab_TabGroup     : 5,
  _ClickTab_ProtectTab   : 6,
  _ClickTab_PinTab       : 7,
  
  _TargetLink_NewTab  : 0,
  _TargetLink_Native  : 1,
  _TargetLink_Current : 2,

  _CloseButtonTMLite    : 4,
  _CloseButtonDefault   : 1,
  _CloseButtonOld       : 3,
  
  numExtTab     : 0,
  numClosedTab  : -1,  
  prefs         : null,
  dragOverDelay : 600,
  unprotectDelay : 300,

  init : function () { 
    window.removeEventListener("load", TabMix.init, false);
    
    TabMix.addMethod();
    TMPrefObserver.register();
    
    if ( !gBrowser )
      gBrowser = document.getElementById("content");
      
    let tabContextMenu = document.getElementById("tabContextMenu");
    let contentMenu    = document.getElementById("contentAreaContextMenu");
    let toolMenu       = document.getElementById("menu_ToolsPopup");
    let appMenu        = document.getElementById("appmenu-popup");
    //let allTabsPopup   = document.getElementById("alltabs-popup");
    let tabContainer   = gBrowser.tabContainer;
 
    gBrowser.addEventListener("click", TabMix.onContentAreaClick, true);
    gBrowser.addTabsProgressListener(TabMixProgressListener);
 
    tabContainer.addEventListener("mousedown", TabMix.onMouseDown, true);
    tabContainer.addEventListener("click", TabMix.onClick, false);
    tabContainer.addEventListener("dblclick", TabMix.onTabBarDblClick, false);
    
    tabContainer.addEventListener("TabOpen", TabMix.onTabOpen, false);
    tabContainer.addEventListener("TabClose", TabMix.onTabClose, false);
    tabContainer.addEventListener("TabSelect", TabMix.onTabSelect, false);
    tabContainer.addEventListener("TabMove", TabMix.onTabMove, false);
    tabContainer.addEventListener("SSTabRestoring", TabMix.onSSTabRestoring, false);
    tabContainer.addEventListener("SSTabRestored", TabMix.onSSTabRestored, false);
    
    tabContainer.addEventListener("dragover", TabMix.onDragOver, false);
    tabContainer.addEventListener("dragend", TabMix.onDragEnd, false);
    tabContainer.addEventListener("drop", TabMix.onDrop, false);
    
    //allTabsPopup.addEventListener("popupshown", TabMix.onAllTabsPopupShown, false);    
    tabContextMenu.addEventListener("popupshowing", TabMix.onTabbarContext, false);
    contentMenu.addEventListener("popupshowing", TabMix.onContentMenu, false);
    toolMenu.addEventListener("popupshowing", TabMix.checkResumeSessionOnceMenu, false);

    if (appMenu)
      appMenu.addEventListener("popupshowing", TabMix.checkResumeSessionOnceMenu, false);
   
    TabMix.prefs = Services.prefs;
    TabMix._persistAttribute();
    TabMix.onUpdate();
    
    if (gBrowser.mCurrentTab) {
      let current = gBrowser.mCurrentTab;
      TabMix.initTab(current);
      if ( current._tPos == 0 )
        current.setAttribute("visited", "true");
    }
    
    // set lock icon for gtk theme
    if ( TabMixLib.setCSS("chrome://browser/skin/browser.css",
                          "#home-button",
                          "list-style-image",
                          null).indexOf("gtk-home") > -1 )                          
      TabMixLib.setCSS("chrome://tabmix/content/tab.css",
                       '.tabbrowser-tab[protected="true"]:not([pinned="true"]) .tab-close-button',
                       "list-style-image",
                       'url("chrome://browser/skin/Secure.png")');
  },
  
  uninit : function () {
    TMPrefObserver.unregister();

    let tabContextMenu = document.getElementById("tabContextMenu");
    let contentMenu    = document.getElementById("contentAreaContextMenu");
    let toolMenu       = document.getElementById("menu_ToolsPopup");
    let appMenu        = document.getElementById("appmenu-popup");
    //let allTabsPopup   = document.getElementById("alltabs-popup");
    let tabContainer   = gBrowser.tabContainer;
 
    gBrowser.removeEventListener("click", TabMix.onContentAreaClick, true);
    gBrowser.removeTabsProgressListener(TabMixProgressListener);

    tabContainer.removeEventListener("mousedown", TabMix.onMouseDown, true);
    tabContainer.removeEventListener("click", TabMix.onClick, false);
    tabContainer.removeEventListener("dblclick", TabMix.onTabBarDblClick, false);
    
    tabContainer.removeEventListener("TabOpen", TabMix.onTabOpen, false);
    tabContainer.removeEventListener("TabClose", TabMix.onTabClose, false);
    tabContainer.removeEventListener("TabSelect", TabMix.onTabSelect, false);
    tabContainer.removeEventListener("TabMove", TabMix.onTabMove, false);
    tabContainer.removeEventListener("SSTabRestoring", TabMix.onSSTabRestoring, false);
    tabContainer.removeEventListener("SSTabRestored", TabMix.onSSTabRestored, false);
    
    tabContainer.removeEventListener("dragover", TabMix.onDragOver, false);
    tabContainer.removeEventListener("dragend", TabMix.onDragEnd, false);
    tabContainer.removeEventListener("drop", TabMix.onDrop, false);
    
    //allTabsPopup.addEventListener("popupshown", TabMix.onAllTabsPopupShown, false);
    tabContextMenu.removeEventListener("popupshowing", TabMix.onTabbarContext, false);
    contentMenu.removeEventListener("popupshowing", TabMix.onContentMenu, false);
    toolMenu.removeEventListener("popupshowing", TabMix.checkResumeSessionOnceMenu, false);

    if (appMenu)
      appMenu.removeEventListener("popupshowing", TabMix.checkResumeSessionOnceMenu, false);
  },
  
  /*
   * Persist tab attributes
   */
  _persistAttribute : function() {
    this._Ss.persistTabAttribute("hidden");
    this._Ss.persistTabAttribute("visited"); 
    this._Ss.persistTabAttribute("protected");
    this._Ss.persistTabAttribute("locked");
  },
  
  /*
   * update the attributes in all Tabs
   */
  onUpdate : function () {
    try {
      var maxWidth       = this.prefs.getIntPref("browser.tabs.tabMaxWidth");
      var minWidth       = this.prefs.getIntPref("browser.tabs.tabMinWidth");
      var related        = this.prefs.getBoolPref("browser.tabs.insertRelatedAfterCurrent");
      var closeButton    = __TM_Prefs.getIntPref("closeButtons");
      var unreadTab      = __TM_Prefs.getBoolPref("unreadTab");
      var group          = __TM_Prefs.getBoolPref("tabGroup");
      var autoLock       = __TM_Prefs.getBoolPref("lockProtectedTab");
    }
    catch(e) {
      var maxWidth       = 250;
      var minWidth       = 100;
      var related        = true;
      var closeButton    = this._CloseButtonDefault;
      var unreadTab      = false;
      var group          = true;
      var autoLock       = false;
    }
    
    var tabs    = gBrowser.tabs;
    var tabBar  = gBrowser.tabContainer;
    //var allTabs = document.getElementById("alltabs-popup");
    
    var TMCloseButton = (closeButton == this._CloseButtonTMLite);
    this.prefs.setIntPref("browser.tabs.closeButtons", TMCloseButton ? this._CloseButtonOld : closeButton);    
    
    tabBar.setAttribute("unread", unreadTab);
    tabBar.setAttribute("tmCloseButton", TMCloseButton);

    // Colorful Tabs Group
    group = group && related;
    TabMixGroup.enabled = group;
    if (group)
      __TM_Prefs.setBoolPref("showMoveToGroup", false);
    let TabViewCommand = document.getElementById("Browser:ToggleTabView");
    let TabViewDeck = document.getElementById("tab-view-deck");
    TabViewCommand.setAttribute("oncommand", group ? "return;" : "TabView.toggle();");
    TabViewDeck.setAttribute("grouping", group);
    tabBar.setAttribute("grouping", group);
    //allTabs.setAttribute("grouping", group);
      
    /* set tab width */
    TabMixLib.setCSS( "chrome://browser/content/browser.css",
                      ".tabbrowser-tab:not([pinned])",
                      "max-width",
                      maxWidth + "px" );
    TabMixLib.setCSS( "chrome://browser/content/browser.css",
                      ".tabbrowser-tab:not([pinned])",
                      "min-width",
                      minWidth + "px" );
    
    /* automatically lock protected tabs */
    let tabs = Array.filter(gBrowser.tabs, function(tab) {
                return tab.hasAttribute("protected");
               });
    Array.forEach( tabs, function(tab) {
      if (autoLock)
        tab.setAttribute("locked", true);
      else
        tab.removeAttribute("locked");
    });
  },
  
  /*
   * General functions
   */
  // tab initialization
  initTab : function (aTab) {
    aTab.groupID = null;
    aTab._opener = null;
  },
  
  // We can't use gBrowser.visibleTabs here, because aTab is removed from the array 
  //   in the case of closing a tab (when gBrowser.removeTab() is called)
  //
  nextTab : function (aTab) {
    let next = null;
    let tab = aTab;
    while ( tab.nextSibling ) {
      tab = tab.nextSibling;
      if ( !tab.hidden ) {
        next = tab;
        break;
      }
    }
    return next;
  },
  
  prevTab : function (aTab) {
    let prev = null;
    let tab = aTab;
    while ( tab.previousSibling ) {
      tab = tab.previousSibling;
      if ( !tab.hidden ) {
        prev = tab;
        break;
      }
    }
    return prev;
  },
  
  /*
   * Link event
   *   Handle the links with a |target| attribute
   */
  _existsFrameName : function (containerFrame, targetFrame) {
    for (var i = 0; i < containerFrame.length; ++i) {
      if (containerFrame[i].name == targetFrame)
        return true;
      if (containerFrame[i].frames.length)
        var return_var = this._existsFrameName(containerFrame[i].frames,targetFrame);
    }
    if (return_var)
      return return_var;

    return false;
  },
  
  onContentAreaClick : function (event) {
    var tmPref = __TM_Prefs.getIntPref("target");
    if ( tmPref == TabMix._TargetLink_Native || !(event.target instanceof HTMLElement) || 
         event.button != 0 || event.ctrlKey || event.shiftKey || event.altKey || event.metaKey )
      return;
    
    var [href, linkNode] = hrefAndLinkNodeForClickEvent(event);
    
    if ( !linkNode )
      return;
  
    let target = linkNode.target;
    if ( target && target != "_content" && target != "_main" &&
         target != "_self" && target != "_parent" && target != "_top" &&             
         !TabMix._existsFrameName(document.commandDispatcher.focusedWindow.top.frames, target) ) {
      if ( tmPref == TabMix._TargetLink_Current )
        linkNode.setAttribute("target", "_top");
      else
        linkNode.setAttribute("target", "_blank");
    }
  },
  
  /*
   * Tab features
   */
  
  // Move Tab
  //   this function is used to move tab without passing onTabMove function
  moveTabTo : function TM_moveTabTo( aTab, aPos ) {
    gBrowser.moveTabTo( aTab, aPos );
  },
  
  // Protect tab
  //   Automatically lock a protected tab
  protectTab : function (aTab) {
    if ( aTab.hasAttribute("protected") ) {
      aTab.removeAttribute("protected");
      aTab.removeAttribute("locked");
    }
    else {
      aTab.setAttribute("protected", "true");
      if ( __TM_Prefs.getBoolPref("lockProtectedTab") )
        aTab.setAttribute("locked", "true");
    }
  },
  
  // Pin or Unpin tab
  pinOrUnpinTab : function (aTab) {
    if ( aTab.hasAttribute("pinned") && aTab.getAttribute("pinned") == "true" )
      gBrowser.unpinTab(aTab);
    else
      gBrowser.pinTab(aTab);
  },
  
  // Duplicate Tab
  //   Use gBrowser.duplicateTab() to duplicate Tab.
  //   Then call gBrowser.moveToTab() to move the new tab next to the original one.
  //
  duplicateTab : function (aTab) {
    var newTab = gBrowser.duplicateTab(aTab);
    gBrowser.moveTabTo(newTab, aTab._tPos+1);
    TabMixGroup.addFromTab(newTab, aTab);
    newTab._opener = aTab._opener;
  },  
    
  // Close Right Tabs an Left Tabs
  //  
  closeMultipleTabs : function (aLeftRight) {
    let tabs = gBrowser.visibleTabs;
    let tab = TabContextMenu.contextTab;
    let index = tabs.indexOf(tab);
    
    let start = ( aLeftRight == "left" ) ? index - 1 : tabs.length - 1;
    let stop  = ( aLeftRight == "left" ) ? gBrowser._numPinnedTabs : index + 1;
    let tabsToClose = start - stop + 1;
    
    if ( tabsToClose < 1 || !gBrowser.TMwarnAboutClosingTabs( false, tabsToClose ) )
      return;

    for ( let i = start; i >= stop; i-- )
      gBrowser.removeTab(tabs[i]);
  },
  
  /*
   * Mouse events :
   *   Click on Tab
   *   Ctrl-click on Tab and Tab Bar
   *   Double-click on Tab
   */
  _tabClickAction : function ( aPref, aTab ) {
    switch (aPref) {
      case this._ClickTab_Null :
        break;
      case this._ClickTab_OpenNewTab :
        BrowserOpenTab();
        break;
      case this._ClickTab_RemoveTab :
        gBrowser.removeCurrentTab();
        break;
      case this._ClickTab_ProtectTab :
        this.protectTab(aTab);
        break;
      case this._ClickTab_PinTab :
        this.pinOrUnpinTab(aTab);
        break;
      case this._ClickTab_DuplicateTab :
        this.duplicateTab(aTab);
        break;
      case this._ClickTab_ReloadTab :
        BrowserReload();
        break;
      case this._ClickTab_TabGroup :
        this.toggleFolder(aTab);
        break
    }
  },
  
  onTabBarDblClick : function (event) {
    if ( event.button != 0 || event.ctrlKey || event.metaKey || event.shiftKey || event.altKey )
      return;
  
    // Double-click on Tab Bar, then do nothing.
    // Firefox will open a new tab itself.
    //
    if ( event.target.localName != "tab" )
      return;
    
    var pref = __TM_Prefs.getIntPref("dblClickTab");
    TabMix._tabClickAction( pref, event.target );
  },
  
  onClick : function (event) {
    var clickOutTabs = event.originalTarget.localName == "box" || event.originalTarget.localName == "scrollbox";
    var clickOnTab = event.target.localName == "tab";
    if ( !clickOutTabs )
      var tab = event.target;

    // Ctrl-click on Tab Bar will undo close tab
    if ( clickOutTabs && event.button == 0 && (event.ctrlKey || event.metaKey) ) {
      undoCloseTab();
      event.stopPropagation();
      event.preventDefault();
    }
    
    // Ctrl-click on a Tab
    if ( clickOnTab && event.button == 0 && (event.ctrlKey || event.metaKey) ) {
      var pref = __TM_Prefs.getIntPref("ctrlClickTab");
      TabMix._tabClickAction( pref, tab );
      event.stopPropagation();
    }
  
    // Alt-click on a Tab
    /* Disabled due to imcompatibility with Firefox button
    if ( clickOnTab && event.button == 0 && event.altKey ) {
      var pref = __TM_Prefs.getIntPref("altClickTab");
      TabMix._tabClickAction( pref, tab );
      event.stopPropagation();
    }*/
  
    // Shift-click on a Tab
    if ( clickOnTab && event.button == 0 && event.shiftKey ) {
      var pref = __TM_Prefs.getIntPref("shiftClickTab");
      TabMix._tabClickAction( pref, tab );
      event.stopPropagation();
    }
  
    // Expand a closed color group
    if ( clickOnTab && event.button == 0 && !event.ctrlKey && !event.metaKey && 
         event.originalTarget.className == "tab-content" &&
         tab.hasAttribute("groupID") && tab.hasAttribute("hidden") ) {
      TabMix.toggleFolder(tab);
      return;
    }

    // Click to select a Tab, see onMouseDown function below
    if ( clickOnTab && event.detail == 1 && event.button == 0 && !event.ctrlKey && !event.metaKey && 
         ( event.originalTarget.className != "tab-close-button" || tab.hasAttribute("protected") ) ) {
      gBrowser.selectedTab = tab;
    }

    // Long click on lock icon to unprotect Tab, see also onMouseDown
    if ( clickOnTab && event.detail == 1 && event.button == 0 && !event.ctrlKey && !event.metaKey && 
         event.originalTarget.className == "tab-close-button" && tab.hasAttribute("protected") ) {
      let time = Date.now();
      if ( tab.readyToUnprotect && time - tab.readyToUnprotect > TabMix.unprotectDelay )
        TabMix.protectTab(tab);
      tab.readyToUnprotect = null;
    }
  },
  
  onMouseDown : function (event) {
    if ( __TM_Prefs.getBoolPref("clickFocus") && event.button == 0 &&
         event.target.localName == "tab" && event.originalTarget.className != "tab-close-button" )
      event.stopPropagation();

    if ( event.button == 0 && !event.ctrlKey && !event.metaKey && 
         event.target.localName == "tab" && event.originalTarget.className == "tab-close-button" &&
         event.target.hasAttribute("protected") ) {
      event.target.readyToUnprotect = Date.now();
    }
  },

  /*
   * Colorful Tab Grouping functions
   *   These are the functions accesible to user
   */
  
  toggleFolder : function (aTab) {
    let ID = aTab.groupID;
    if ( !ID )
      return;

    if ( aTab.hasAttribute("hidden") )
      TabMixGroup.openStack(ID);
    else
      TabMixGroup.makeStack(aTab);
  },
  
  bookmarkGroup : function (aTab) {
    let tabs = TabMixGroup.getTabsInGroup(aTab);
    let URIs = tabs.map( function __getUri(tab) {
                 return tab.linkedBrowser.webNavigation.currentURI;
               });
    PlacesUIUtils.showMinimalAddMultiBookmarkUI(URIs);
  },
  
  reloadGroup : function (aTab) {
    let tabs = TabMixGroup.getTabsInGroup(aTab);
    for ( var i = 0; i < tabs.length; i++ ) {
      try {
        gBrowser.getBrowserForTab(tabs[i]).reload();
      }
      catch (e) {}
    }
  },
  
  groupOrphanTabs : function (aTab) {
    let tabs = gBrowser.visibleTabs;
    let tab = aTab.nextSibling;
    while ( tab && !tab.hasAttribute("groupID") ) {
      if ( tabs.indexOf(tab) >= 0 && !tab.pinned ) // make sure |tab| is in the current Tabview group
        TabMixGroup.addFromTab( tab, aTab );
      tab = tab.nextSibling;
    }
    tab = aTab.previousSibling;
    while ( tab && !tab.hasAttribute("groupID") ) {
      if ( tabs.indexOf(tab) >= 0 && !tab.pinned )
        TabMixGroup.addFromTab( tab, aTab );
      tab = tab.previousSibling;
    }
  },
  
  // Remove one tab from its group
  //
  removeFromGroup : function TM_removeFromGroup (aTab) {
    let ID = aTab.groupID;
    if ( ID == null )
      return;
    TabMixGroup.removeOneTab(aTab);
    
    // Then we need to move aTab to the end of the group
    let tabs = TabMixGroup.getTabsInGroup(ID);
    if ( tabs.length > 0 && aTab._tPos > tabs[0]._tPos && aTab._tPos < tabs[tabs.length - 1]._tPos )
      this.moveTabTo( aTab, tabs[tabs.length - 1]._tPos );
  },
  
  undoGroup : function (aTab) {
    let tabs = TabMixGroup.getTabsInGroup(aTab);
    
    // the tabs[0] will automatically removed when we remove the last one, see TabMixGroup.removeOneTab()
    for ( let i = 1; i < tabs.length; i++ ) { 
      TabMixGroup.removeOneTab(tabs[i]);
      tabs[i]._opener = null;
    }
    tabs[0]._opener = null;
  },
  
  closeAllTabsInGroup : function (aTab) {
    let tabs = TabMixGroup.getTabsInGroup(aTab);
    let last = null;
    
    let tabToClose = tabs.length;
    if ( gBrowser.TMwarnAboutClosingTabs(false, tabToClose) ) {
      for ( let i = tabs.length - 1; i >= 0; i-- ) {
        // if the current tab is in the group, we remove it last
        if ( tabs[i] != gBrowser.mCurrentTab )
          gBrowser.removeTab(tabs[i]);
        else
          last = tabs[i];
      }
      if ( last )
        gBrowser.removeTab(last);
    }
  },

  _clearOpener : function () {
    Array.forEach( gBrowser.visibleTabs, function (tab) {
    if ( !tab.groupID )
      tab._opener = null;
    });
  },

  /*
   * Drag and drop to group orphan tabs
   */
  onDragOver : function (event) {
    if ( !TabMixGroup.enabled )
      return;

    let tab = event.target;
    let draggedTab = event.dataTransfer.mozGetDataAt(TAB_DROP_TYPE, 0);
    if ( tab.localName == "tab" && event.dataTransfer.dropEffect == "move" && tab != draggedTab ) {
      let boxObject = tab.boxObject;
      if ( event.screenX > boxObject.screenX + boxObject.width * .25 &&
           event.screenX < boxObject.screenX + boxObject.width * .75 ) {
        // this = tabContainer
        if ( !this._dragTime )
          this._dragTime = Date.now();
        if ( Date.now() >= this._dragTime + TabMix.dragOverDelay && !tab.hasAttribute("TMtoMakeGroup") ) {
          let color = TabMixGroup.pickColor(tab);
          tab.setAttribute( "TMtoMakeGroup", color );
          draggedTab.setAttribute( "TMtoMakeGroup", color );
        }
      }
      else {
        tab.removeAttribute("TMtoMakeGroup");
        draggedTab.removeAttribute("TMtoMakeGroup");
      }
    }
  },
  
  onDragEnd : function (event) {
    Array.forEach( gBrowser.tabs, function (tab) {
      tab.removeAttribute("TMtoMakeGroup");
    });
  },

  onDrop : function (event) {
    if ( !TabMixGroup.enabled )
      return;
    
    let dt = event.dataTransfer;

    if ( dt.dropEffect != "move" )
      return;
    
    let draggedTab = dt.mozGetDataAt(TAB_DROP_TYPE, 0);
    if (!draggedTab)
      return;
    
    let targetTab = event.target;
    
    if ( targetTab.hasAttribute("TMtoMakeGroup") ) {
      if ( draggedTab.groupID )
        TabMixGroup.removeOneTab(draggedTab);
      targetTab.removeAttribute("TMtoMakeGroup");
      draggedTab.removeAttribute("TMtoMakeGroup");
      TabMixGroup.addFromTab( draggedTab, targetTab );
    }
  },

  /*
   * Tab events
   */

  // Regroup tabs
  //
  //  In any of the cases below, we don't do anything
  //   1. when Firefox move a related tab next to the opener
  //   2. when restoring a closed tab, see below
  //   3. when Tab Mix Lite tries to move a tab to its correct position
  //
  onTabMove : function (event) {
    if ( !TabMixGroup.enabled )
      return;
      
    let tab = event.target;
    
    // when restoring a closed tab, onTabMove() will be called firstly by Firefox restore function,
    // and we have to call TabMixGroup.moveRestoredTab() here, because onSSTabRestored() will only
    // be called after the restored page is finished loading.
    if ( tab._restoring ) {
      TabMixGroup.moveRestoredTab(tab);
      return;
    }

    // what function have called gBrowser.moveTabTo() ?
    // |caller| is no-standard but useful
    try {
      let callerName = arguments.callee.caller.caller.name;
      if ( callerName == "addTab" || 
           callerName == "TM_moveTabTo" )
        return;
    }
    catch (e) { 
      TabMixLib.msg("onTabMove can't get caller"); // would be the case of tabviewhide
      //return;
    }

    let nextTab = TabMix.nextTab(tab);
    let prevTab = TabMix.prevTab(tab);

    let oID    = tab.groupID;
    let nextID = nextTab == null ? null : nextTab.groupID;
    let prevID = prevTab == null ? null : prevTab.groupID;
    
    // when a tab stack is moved, all the tabs should be moved together
    //
    // when moving the other tabs of the stack, onTabMove() will be called again
    // so we need to set TabMixGroup.moving to true, to prevent an infinited loop
    // see TabMixGroup.moveStack()
    //
    // event.detail is the tab's original position
    //
    if ( tab.hasAttribute("hidden") && !TabMixGroup.moving ) {
      TabMixGroup.moveStack(tab, event.detail);
      return;
    }
    
    // the moved tab is dropped before a tab stack
    // see TabMixGroup.tabBeforeStack()
    if ( nextID && nextTab.hasAttribute("hidden") && nextTab.getAttribute("hidden") != "true" )
      TabMixGroup.tabBeforeStack(tab, nextTab);
    
    // still in its own group, or the moved tab and its neighbors are all orphan tabs
    if ( oID == nextID || oID == prevID ) {
      let prevParent = prevTab ? ( null || prevTab._opener ) : null;
      let nextParent = nextTab ? ( null || nextTab._opener ) : null;
      if ( oID && prevTab && nextTab && prevParent == nextParent )
        tab._opener = nextTab._opener;
      return;
    }

    // outside its original group
    if ( oID && !TabMixGroup.moving )
      TabMixGroup.removeOneTab(tab);
    
    // into another group 
    if ( nextID && nextID == prevID ) {
      TabMixGroup.setID(tab, nextID);
      tab.removeAttribute("hidden");
      tab._opener = nextTab._opener;
    }
    // new position has no group
    else {
      if ( oID )
        tab._opener = null;
    }
  },
  
  // When Tab is selected, set |visited|
  //
  onTabSelect : function (event) {
    TabMix.numPinTab = 0;
    var tab = event.target;
    tab.setAttribute("visited", true);
  },
  
  // "Smart focus" after closing a tab
  //
  onTabClose : function (event) {
    TabMix.numClosedTab++;

    if ( !TabMixGroup.enabled || !TabMix.prefs.getBoolPref("browser.tabs.selectOwnerOnClose") )
      return;
      
    let tab = event.target;
    let group = tab.groupID;
    let isCurrent = tab.hasAttribute("selected");
    
    let target = null || tab._opener;
    let newTab = null;
    let nextTab = TabMix.nextTab(tab);
    let prevTab = TabMix.prevTab(tab);


    if (group) {
      if ( target && isCurrent ) {
        let tabs = TabMixGroup.getTabsInGroup(group);
        let inside = tabs.indexOf(target) >= 0; // whether the opener is in the same group

        Array.forEach( tabs, function (tempTab) {
          // pass the opener to its children
          // and also to its orphan siblings if the opener is outside the group
          if ( tempTab._opener == tab || ( tempTab._opener == null && !inside ) )
            tempTab._opener = target;
        });

        if ( nextTab && nextTab.groupID == group && nextTab._opener == target )
          newTab = nextTab;
        else if ( prevTab && prevTab.groupID == group && prevTab._opener == target )
          newTab = prevTab;
        else if (inside) // opener inside the group, no siblings, go back to its opener
          newTab = target;
        // opener outside the group, no siblings, newTab = null
      }
      if ( !newTab && isCurrent ) {
        if ( nextTab && nextTab.groupID == group )
          newTab = nextTab;
        else if ( prevTab && prevTab.groupID == group )
          newTab = prevTab;
        // error (siblings not found), newTab = null
      }
      TabMixGroup.removeOneTab(tab);
    }
    else {
      let orphan = false;
      Array.forEach( gBrowser.visibleTabs, function (tempTab) {
        if ( tempTab._opener == tab )
          orphan = true;
      });
      if (orphan) {
        TabMix._clearOpener();
        return;
      }
      if (target)
        newTab = target;
    }

    // it's a background tab we don't need to select other tab
    if ( !isCurrent )
      return;

    if ( newTab && gBrowser.visibleTabs.indexOf(newTab) >= 0 ) {
      gBrowser.selectedTab = newTab;      
      newTab.setAttribute("visited", true);
    }
  },
  
  // When opening a tab, check if it's in a tab group or a popup
  //
  onTabOpen : function (event) {
    var tab = event.target;
    TabMix.initTab(tab);
    
    if ( !TabMixGroup.enabled )
      return;
        
    let current = gBrowser.mCurrentTab;
    
    if ( current.pinned ) {
      if ( tab.related && gBrowser._lastRelatedTab )
        TabMixGroup.addFromTab(tab, gBrowser.tabs[gBrowser._numPinnedTabs]);
      tab._opener = current;
    }
    else {
      if ( gBrowser._lastRelatedTab == null ) {
        current._allowOpener = true;
        let tabs = TabMixGroup.getTabsInGroup(current);
        Array.forEach( tabs, function (tempTab) {
          if ( tempTab._opener == current )
            tempTab._opener = null;
        });        
      }
        
       // new group to be created
      if ( tab.related && current.groupID == null ) {
        current._allowOpener = false;
        // we have already some related tabs opened, but these tabs are not grouped
        if ( gBrowser._lastRelatedTab )
          gBrowser._lastRelatedTab = null;
      }

      // numExtTab: tabs opened from external app
      if ( tab.related || TabMix.numExtTab > 1 )
        TabMixGroup.addFromTab(tab, current);
      
      if ( tab.owner || ( tab.related && current._allowOpener ) )
        tab._opener = current;
    }
  },
  
  // Don't do anything when restoring a session
  onSSTabRestoring : function (event) {
    // hack to know whether we are reopening a browser or calling undoclosetab
    let realCount = TabMix._Ss.getClosedTabCount(window);
    let tempCount = TabMix.numClosedTab;
    
    let tab = event.target;
    if ( tab.hasAttribute("visited") )
      tab.visited = true;
    if ( tab.hasAttribute("groupID") )
      tab.groupID = tab.getAttribute("groupID");
/*
    if ( tempCount == realCount ) // session restoring
      return;
    else { // undoclosetab when tempCount != -1
      TabMix.numClosedTab = realCount;
      if ( tempCount == -1 ) // browser is just opened
        return;
    }
*/  
    TabMixGroup.restore(tab);
  },
  
  onSSTabRestored : function (event) { 
    if ( event.target._restoring )
      TabMixGroup.moveRestoredTab(event.target);
  },

  onWindowActive : function () {
    TabMix.numExtTab = 0;
  },
  
  // override gBrowser.loadTabs
  //   set |owner| of new tabs
  //   reuse the current blank tab
  //   set tabs' group
  //
  loadTabs : function ( aURIs, aLoadInBackground, aReplace ) {
    if (!aURIs.length)
      return;

    var owner = aLoadInBackground ? null : gBrowser.selectedTab;
    aReplace = aReplace || ( gBrowser.mCurrentBrowser.webNavigation.currentURI.spec == "about:blank" &&
                             !gBrowser.mCurrentTab.hasAttribute("busy") );
    var firstTabAdded = null;

    if (aReplace) {
      try {
        gBrowser.loadURI(aURIs[0], null, null);
      } catch (e) {
      }
    }
    else
      firstTabAdded = gBrowser.addTab(aURIs[0], {ownerTab: owner});

    if (!aLoadInBackground) {
      if (firstTabAdded) {
        gBrowser.selectedTab = firstTabAdded;
      }
      else
        window.content.focus();
    }

    for (let i = 1; i < aURIs.length; ++i) {
      let tab = gBrowser.addTab(aURIs[i], {relatedToCurrent: true});
    }
  },

  /*
   * Keyboard event :
   *   Alt-C will toggle the focus between the current tab and the page content
   */
  toggleFocus : function () {
    if ( !gBrowser.mStrip.collapsed && document.commandDispatcher.focusedElement != gBrowser.mCurrentTab )
      gBrowser.mCurrentTab.focus();
    else
      window.content.focus();
  },

  /*
   * Menu events :
   *   "Restore only the current session" on Tools Menu
   *   Create menuitems for Tab Bar Context Menu
   *   On Tab Bar Context Menu showing
   *   On Content Area Menu showing
   */
   
  // Tools Menu : Restore only the current session
  //
  toggleResumeSessionOnce : function () {
    var pref = "browser.sessionstore.resume_session_once";
    this.prefs.setBoolPref(pref, !getBoolPref(pref));
  },
  
  // Check the this menuitem when opening the Tool Menu
  //
  checkResumeSessionOnceMenu : function () {
    var toolSR = document.getElementById("tm-tool-restoreSessionOnce");
    var appSR  = document.getElementById("tm-appmenu-restoreSessionOnce");
    var __PBS = Components.classes["@mozilla.org/privatebrowsing;1"]
                          .getService(Components.interfaces.nsIPrivateBrowsingService); 
    
    toolSR.hidden = !__TM_Prefs.getBoolPref("toolsSR") || __PBS.privateBrowsingEnabled;
    toolSR.setAttribute("checked", getBoolPref("browser.sessionstore.resume_session_once") );
    
    if ( appSR ) {
      appSR.hidden  = !__TM_Prefs.getBoolPref("toolsSR") || __PBS.privateBrowsingEnabled;
      appSR.setAttribute("checked", getBoolPref("browser.sessionstore.resume_session_once") );
    }
  },
  
  // On Content Area Menu showing
  //
  onContentMenu : function () {
    var showUndoCloseTab  = __TM_Prefs.getBoolPref("undoCloseTabContent");
    var showUndoCloseList = __TM_Prefs.getBoolPref("undoCloseListContent");

    var closedTabCount    = TabMix._Ss.getClosedTabCount(window);
    var undoCloseTabMenu  = document.getElementById("tm-content-undoCloseTab");

    var undoCloseListMenu = document.getElementById("tm-content-undoCloseList");
    var undoCloseSep      = document.getElementById("tm-content-undoCloseSep");
    var contentClick      = gContextMenu.isTextSelected || gContextMenu.onTextInput ||
                            gContextMenu.onLink || gContextMenu.onImage;

    undoCloseTabMenu.hidden = !closedTabCount || !showUndoCloseTab || contentClick;
    undoCloseListMenu.hidden = !closedTabCount || !showUndoCloseList || contentClick;
    undoCloseSep.hidden = undoCloseTabMenu.hidden && undoCloseListMenu.hidden;
  },
  
  // On Tab Bar Context Menu showing
  //
  onTabbarContext : function () {
    let closedTabCount    = TabMix._Ss.getClosedTabCount(window);
    let duplicateMenu     = document.getElementById("tm-duplicateTab");
    let protectMenu       = document.getElementById("tm-protectTab");
    let tabGroupMenu      = document.getElementById("tm-tabGroup");
    let undoCloseListMenu = document.getElementById("tm-undoCloseList");
    let closeRightMenu    = document.getElementById("tm-closeRightTabs");
    let closeLeftMenu     = document.getElementById("tm-closeLeftTabs");
    //let closeSepMenu      = document.getElementById("tm-separator");
    let pinTab            = document.getElementById("context_pinTab");
    let moveToGroupMenu   = document.getElementById("context_tabViewMenu");
    let moveToNewWindow   = document.getElementById("context_openTabInWindow");

    let tab     = TabContextMenu.contextTab;
    let nextTab = TabMix.nextTab(tab);
    let prevTab = TabMix.prevTab(tab);
    
    let firstOrLastTab = ( !nextTab || !prevTab );
    
    duplicateMenu.hidden     = !__TM_Prefs.getBoolPref("duplicateMenu");
    protectMenu.hidden       = !__TM_Prefs.getBoolPref("protectMenu");
    moveToGroupMenu.hidden   = !__TM_Prefs.getBoolPref("showMoveToGroup");
    moveToNewWindow.hidden   = !__TM_Prefs.getBoolPref("showMoveToNewWindow");
    closeRightMenu.hidden    = !__TM_Prefs.getBoolPref("closeRightMenu") || firstOrLastTab;
    closeLeftMenu.hidden     = !__TM_Prefs.getBoolPref("closeLeftMenu") || firstOrLastTab;
    undoCloseListMenu.hidden = !__TM_Prefs.getBoolPref("undoCloseListMenu") || !closedTabCount;
    //closeSepMenu.hidden      = closeRightMenu.hidden && closeLeftMenu.hidden;
    
    if ( tab.hasAttribute("hidden") )
      pinTab.disabled = true;
    
    var singleOrphan = !tab.hasAttribute("groupID") &&
                       ( !nextTab || nextTab.hasAttribute("groupID") ) &&
                       ( !prevTab || prevTab.hasAttribute("groupID") );
    tabGroupMenu.hidden = !__TM_Prefs.getBoolPref("tabGroup") || singleOrphan || tab.pinned;
    
    if ( tab.getAttribute("protected") == "true" )
      protectMenu.setAttribute("checked", "true");
    else
      protectMenu.removeAttribute("checked");
  },
  
  // On Tab Grouping Menu showing
  //
  onTabGroupMenu : function () {
    var stackGroup      = document.getElementById("tm-groupFolder");
    var bookmarkGroup   = document.getElementById("tm-bookmarkGroup");
    var reloadGroup     = document.getElementById("tm-reloadGroup");
    var removeFromGroup = document.getElementById("tm-removeFromGroup");
    var groupSep        = document.getElementById("tm-context-groupSep");
    var undoGroup       = document.getElementById("tm-undoGroup");
    var closeGroup      = document.getElementById("tm-closeGroup");
    var groupOrphans    = document.getElementById("tm-groupOrphans");
    
    var stacked = TabContextMenu.contextTab.hasAttribute("hidden");
    var grouped = TabContextMenu.contextTab.hasAttribute("groupID");
    groupOrphans.hidden    = grouped;
    
    stackGroup.hidden      = !grouped;
    bookmarkGroup.hidden   = !grouped;
    
    removeFromGroup.hidden = stacked || !grouped;
    reloadGroup.hidden     = stacked || !grouped;
    groupSep.hidden        = stacked || !grouped;
    undoGroup.hidden       = stacked || !grouped;
    closeGroup.hidden      = stacked || !grouped;
  },

  // Colorize the grouped tabs
  //
  onAllTabsPopupShown : function (event) {
    if ( !event.isTrusted )
      return;

    for ( let i = 0; i < this.childNodes.length; i++ ) {
      if ( this.childNodes[i].tab ) {
        let tab = this.childNodes[i].tab;
        if ( tab.hasAttribute("groupID") )
          this.childNodes[i].setAttribute( "groupID", tab.getAttribute("groupID") );
      }
    }
  },
  
  /*
   * Closed Tabs List :
   *   Create the Closed Tabs List Menu
   *   Clear all closed tabs
   */
  createUndoCloseList : function (popup) {
    while(popup.hasChildNodes()) {
      popup.removeChild(popup.lastChild);
    }
    
    var undoItems = JSON.parse(TabMix._Ss.getClosedTabData(window));
    for (var i = 0; i < undoItems.length; i++) {
      var m = popup.appendChild(document.createElement("menuitem"));
      m.setAttribute("class", "menuitem-iconic bookmark-item");
      m.setAttribute("label", i + ": " + undoItems[i].title);
      m.setAttribute("accesskey", i.toString());
      m.setAttribute("image", undoItems[i].image);
      m.setAttribute("oncommand", "undoCloseTab(" + i + ");");
    }
  },

  clearAllClosedTabs : function () {
    var max = this.prefs.getIntPref("browser.sessionstore.max_tabs_undo");
    this.prefs.setIntPref("browser.sessionstore.max_tabs_undo",0);
    this.prefs.setIntPref("browser.sessionstore.max_tabs_undo",max);
  },
   
  addMethod : function () {
    // For Locked Tab
    //   code from extension Tab Utilities
    //

    TabMixLib.codeHook(
      "gBrowser.loadURI", 
      "{", 
      ( 'if (!aURI.match(/^javascript:/) && this.selectedTab.hasAttribute("locked")) {' +
        '  return this.loadOneTab(aURI, aReferrerURI, aCharset, null, null, false); }' )
    );

    TabMixLib.codeHook(
      "gBrowser.loadURIWithFlags",
      "{", 
      ( 'if (!aURI.match(/^javascript:/) && this.selectedTab.hasAttribute("locked")) {' +
        ' return this.loadOneTab(aURI, aReferrerURI, aCharset, aPostData, null, aFlags &'+
        '                        Ci.nsIWebNavigation.LOAD_FLAGS_ALLOW_THIRD_PARTY_FIXUP); }' )
    );

    TabMixLib.codeHook(
      "contentAreaClick", 
      /(\}\s*handleLinkClick)/, 
      ( 'if ( gBrowser.selectedTab.hasAttribute("locked") && gBrowser.selectedTab.getAttribute("locked") == "true" ) {' +
        '  openNewTabWith(href, event.target.ownerDocument, null, event, false);' +
        '  event.preventDefault();' +
        '  return true;' +
        '}' +
        '$&' )
    );

/*  backup for future use
    TabMixLib.codeHook(
      "XULBrowserWindow.onBeforeLinkTraversal",
      "{",
      'if ( gBrowser.selectedTab.hasAttribute("locked") ) return "_blank";'
    );
    
    TabMixLib.codeHook(
      "openLinkIn",
      "switch",
      ( 'if ( w.gBrowser.selectedTab.hasAttribute("locked") ) {' +
        '  where = "tab";' +
        '  loadInBackground = false; }' +
        '$&' )
    );
*/    

    // gBrowser.addTab
    //   check if the new tab is a related one
    //   we must do this before dispatching the "TabOpen" event
    //
    TabMixLib.codeHook(
      "gBrowser.addTab",
      't.className = "tabbrowser-tab";',
      ( '$&'+
        't.related = (aRelatedToCurrent == null ? aReferrerURI : aRelatedToCurrent);' )
    );
    
    // gBrowser.removedTab
    //   don't close protected or hidden tab
    //
    TabMixLib.codeHook(
      "gBrowser.removeTab",
      '{',
      'if ( aTab.getAttribute("protected") == "true" || aTab.hasAttribute("hidden") )'+
      '  return;'
    );
    
    // gBrowser.loadTabs
    //   override the original code
    //
    TabMixLib.codeHook(
      "gBrowser.loadTabs",
      '{',
      ( 'if (TabMixGroup.enabled) {'+
        '  TabMix.loadTabs(aURIs, aLoadInBackground, aReplace);'+
        '  return; }' )
    );
    
    // gBrowser.pinTab
    //   a pinned tab should be removed from its group (if it has)
    //
    TabMixLib.codeHook(
      "gBrowser.pinTab",
      '{',
      'if ( aTab.hasAttribute("hidden") ) {' +
      '  return; }' +
      'if ( aTab.hasAttribute("groupID") ) {' +
      '  TabMixGroup.removeOneTab(aTab); }'
    );
  
    // gBrowser.TMwarnAboutClosingTabs
    //   send the number of tabs to be closed when closing left or right tabs or tabs in a group
    //   

    gBrowser.TMwarnAboutClosingTabs = gBrowser.warnAboutClosingTabs;
    
    TabMixLib.codeHook(
      "gBrowser.TMwarnAboutClosingTabs",
      '(aAll)',
      '( aAll, aTabsToClose )'
    );
    
    TabMixLib.codeHook(
      "gBrowser.TMwarnAboutClosingTabs",
      'tabsToClose <= 1',
      'aTabsToClose <= 1'
    );
    
    TabMixLib.codeHook(
      "gBrowser.TMwarnAboutClosingTabs",
      '[tabsToClose]',
      '[aTabsToClose]'
    );

    // Search bar
    //   Override the click behaviour
    //
    TabMixLib.codeHook(
      "document.getElementById('searchbar').handleSearchCommand",
      'where = whereToOpenLink(aEvent, false, true);',
      ( '$&'+
        'var newTabPref = textBox._prefBranch.getBoolPref("browser.search.openintab");'+
        'if ( newTabPref ) {'+
        '  if ( where == "tab" )'+
        '    where = "current";'+
        '  else if ( where == "current" )'+
        '    where = "tab";'+
        '}' )
     );
    
    // When an external application intents to open a link in new tab,
    // keep the Firefox window in background.
    // If the current tab is blank, reuse this tab instead of opening a new one.
    //
    TabMixLib.codeHook(
      "nsBrowserAccess.prototype.openURI",
      'let tab',
      ( 'let tab;' +
        'var tmLoadInBackgroundWin = __TM_Prefs.getBoolPref("openExternalBackground");'+
        'if ( tmLoadInBackgroundWin && isExternal && !loadInBackground ) {'+
        '  needToFocusWin = false;'+
        '  TabMixLib.notification();'+
        '  if ( (++TabMix.numExtTab) > 1 )'+
        '    loadInBackground = true;'+
        '}'+
        'if ( isTabEmpty(win.gBrowser.mCurrentTab) ) {'+
        '  tab = win.gBrowser.mCurrentTab;'+
        '  win.gBrowser.selectedBrowser.focus();'+
        '  if (aURI) {'+
        '    let referrer = aOpener ? makeURI(aOpener.location.href) : null;'+
        '    let loadflags = isExternal ?'+
        '                      Ci.nsIWebNavigation.LOAD_FLAGS_FROM_EXTERNAL :'+
        '                      Ci.nsIWebNavigation.LOAD_FLAGS_NONE;'+
        '    win.gBrowser.loadURIWithFlags(aURI.spec, loadflags, referrer, null, null);'+
        '  }'+
        '}'+
        'else'+
        '  tab' )
    );
    
    TabMixLib.codeHook(
      "nsBrowserAccess.prototype.openURI",
      'if (needToFocusWin || !loadInBackground && isExternal)',
      'if ( needToFocusWin || ( !loadInBackground && isExternal && !tmLoadInBackgroundWin ) )'
    );
  }

}; // End of TabMix

var TabMixProgressListener =
{
  // For unread tab
  //
  onStateChange : function ( aBrowser, aWebProgress, aRequest, aStateFlags, aStatus ) {  
    if (!aRequest)
      return;
    const nsIWebProgressListener = Components.interfaces.nsIWebProgressListener;

    if ( aStateFlags & nsIWebProgressListener.STATE_STOP &&
         aStateFlags & nsIWebProgressListener.STATE_IS_NETWORK) {
      for ( let i = 0; i < gBrowser.tabs.length; i++ ) {
        let tab = gBrowser.tabs[i];
        if ( tab.linkedBrowser == aBrowser ) {
          if ( tab.hasAttribute("visited") && !tab.hasAttribute("selected") ) {
            if ( !tab.visited )
              tab.removeAttribute("visited");
            tab.visited = null;
          }
          break;
        }
      }
    }
  }
};

window.addEventListener("load", TabMix.init, false);
window.addEventListener("unload", TabMix.uninit, false);
window.addEventListener("activate", TabMix.onWindowActive, false);
